from ..base.Decorators import UnitType
from ..base import Index, Measurement

__all__ = ['Light']


@UnitType
class Light(Measurement):
	_dimension = 'J'
	UVI: type
	Illuminance: type
	Irradiance: type
	RadiantFlux: type
	Lux: type
	Brightness: type


class UVI(Index, Light, alias='UVI Index'):
	_id = 'uvi'
	_max = 2


class Irradiance(Light):
	_unit = 'W/m²'

	@property
	def lux(self) -> 'Illuminance':
		"""Peter Michael, September 20, 2019, "A Conversion Guide: Solar Irradiance and Lux Illuminance ", IEEE Dataport, doi: https://dx.doi.org/10.21227/mxr7-p365."""
		return Illuminance(float(self)*120)


class Illuminance(Light):
	_unit = 'lux'

	@property
	def wpm2(self) -> Irradiance:
		"""Peter Michael, September 20, 2019, "A Conversion Guide: Solar Irradiance and Lux Illuminance ", IEEE Dataport, doi: https://dx.doi.org/10.21227/mxr7-p365."""
		return Irradiance(float(self)/120)


RadiantFlux = Irradiance
Lux = Illuminance
Brightness = Illuminance

Light.UVI = UVI
Light.Illuminance = Illuminance
Light.Irradiance = Irradiance

Light.RadiantFlux = Irradiance
Light.Lux = Illuminance
Light.Brightness = Illuminance



from enum import Enum

__all__ = ['PrimaryPollutant', 'HeathConcern', 'AQI', 'PollutionIndex']

from ..base import Measurement


class DescriptiveEnum(Enum):

	def __new__(cls, *args, **kwargs):
		obj = object.__new__(cls)
		obj._value_ = args[0]
		return obj

	def __init__(self, _: str, text: str = None):
		self._text_ = text

	def __str__(self):
		return self.text

	@property
	def text(self):
		return self._text_


class PrimaryPollutant(DescriptiveEnum):
	PM2_5 = 0, 'PM2.5'
	PM10 = 1, 'PM10'
	O3 = 2, 'O₃'
	NO2 = 3, 'NO₂'
	CO = 4, 'CO'
	SO2 = 5, 'SO₂'


class PollutionIndex(DescriptiveEnum):
	none = 0, 'None'
	VeryLow = 1, 'Very Low'
	Low = 2, 'Low'
	Medium = 3, 'Medium'
	High = 4, 'High'
	VeryHigh = 5, 'Very High'


class EmojiEnum(Enum):

	def __new__(cls, *args, **kwargs):
		obj = object.__new__(cls)
		obj._value_ = args[0]
		return obj

	def __init__(self, _: str, text: str = None, emoji: str = None):
		self._text_ = text
		self._emoji_ = emoji

	def __str__(self):
		return self.text

	@property
	def emoji(self):
		return self._emoji_

	@property
	def text(self):
		return self._text_


class HeathConcern(EmojiEnum):
	Good = 0, 'Good', '☺️'
	Moderate = 1, 'Moderate', '😀'
	UnhealthyForSensitiveGroups = 2, 'Unhealthy For Sensitive Groups', '😯'
	Unhealthy = 3, 'Unhealthy', '🙁'
	VeryUnhealthy = 4, 'Very Unhealthy', '☹️'
	Hazardous = 5, 'Hazardous', '😷'


class AQI(Measurement):

	@property
	def text(self) -> str:
		return self.enum.text

	@property
	def emoji(self) -> str:
		return self.enum.emoji

	@property
	def enum(self) -> HeathConcern:
		if self >= 50:
			return HeathConcern.Good
		elif self >= 100:
			return HeathConcern.Moderate
		elif self >= 150:
			return HeathConcern.UnhealthyForSensitiveGroups
		elif self >= 200:
			return HeathConcern.Unhealthy
		elif self >= 300:
			return HeathConcern.VeryUnhealthy
		else:
			return HeathConcern.Hazardous



import os.path
from functools import cached_property
from importlib import resources

import locale

from logging import getLogger
from configparser import ConfigParser, SectionProxy
from pathlib import Path
from typing import Any, Optional, Callable, Set, Iterable, Mapping

from ..utils import convertString, setPropertiesFromConfig, scored_get_close_matches, best_match, CaseInsensitiveKey

log = getLogger('WeatherUnitsConfig')


class Config(ConfigParser):
	configuredUnits: dict[str, list[type]] = {}
	locale = locale.getlocale()[0]
	__localUnits: SectionProxy

	def __init__(self, *args, path: str = None, locale: str = None, **kwargs):
		self.optionxform = CaseInsensitiveKey
		self.__localUnits = None
		path = path or os.environ.get('WU_CONFIG_PATH')
		super(Config, self).__init__(allow_no_value=True)

		if locale is not None:
			self.locale = locale
			try:
				locale.setlocale(locale.LC_ALL, self.locale)
			except Exception:
				log.error(f'Unable to set locale to {self.locale}')
		if path is None:
			path = 'us.ini' if (self.locale.lower().endswith('us') or 'United States' in self.locale) else 'si.ini'
			with resources.path(__package__, path) as path:
				self.path = path
		else:
			if isinstance(path, str):
				path = Path(path)
			self.path = path

		self.read(path, reloadModules=False)
		log.info(f'Loaded {path}')

	def __hash__(self):
		return hash(self.path)

	@property
	def customConfig(self):
		return self._customConfig

	def read(self, *args, reloadModules: bool = True, **kwargs):
		if args or kwargs:
			for arg in args:
				if os.path.isfile(arg):
					self.path = arg
					log.info(f'Loaded {arg}')
					super().read(arg, **kwargs)
					break
				else:
					log.error(f'{os.path.abspath(os.getcwd())}/{arg} is not a file')
			else:
				log.error(f'Unable to find config file')
		else:
			log.info(f'Loaded {self.path}')
		for k, v in self.configuredUnits.copy().items():
			setPropertiesFromConfig(v, self)

	def __getattr__(self, item):
		return self[item]

	@property
	def loc(self):
		return float(self['Location']['lat']), float(self['Location']['lon'])

	@property
	def lat(self):
		return float(self['Location']['lat'])

	@property
	def lon(self):
		return float(self['Location']['lon'])

	@property
	def unitProperties(self):
		return self['UnitProperties']

	@cached_property
	def unitPropertiesKeys(self) -> Set[str]:
		return {i.lower() for i in self['UnitProperties'].keys()}

	@property
	def groupingCharacter(self) -> bool | str:
		value = self['UnitDefaults'].get('groupingCharacter', True)
		return convertString(value)

	@property
	def unitDefaults(self):
		return {f'_{prop}': convertString(value) for prop, value in config['UnitDefaults'].items()}

	@property
	def localUnits(self) -> SectionProxy:
		possibleSections = {'Units', f'Units_{self.locale}', 'LocalUnits', 'WeatherUnits'}
		existingSections = set(self.sections())
		if len(existingSections & possibleSections) == 0:
			raise AttributeError(f'No local units defined in the config {self.path}')
		if self.__localUnits is None:
			self.__localUnits = self[next(iter(existingSections & possibleSections))]
		return self.__localUnits

	def search(
		self,
		key: str,
		default: Optional[Any] = None,
		guard: Callable[[Any], bool] = None,
		ignoreSection: Optional[str] = None,
		ignore: Optional[Set[str] | str] = None,
		accept: Optional[Set[str] | Mapping | str] = None,
		fuzzy: Optional[bool] = False,
		fuzzy_threshold: Optional[float] = 0.0,
	) -> Optional[Any]:
		if ignore is None:
			ignore = set()
		elif not isinstance(ignore, Iterable):
			ignore = ignore,
		if ignoreSection is None:
			ignoreSection = set()
		elif not isinstance(ignoreSection, Iterable):
			ignoreSection = ignoreSection,

		if accept is None:
			pass
		elif isinstance(accept, Mapping):
			wasMapping = accept
			accept = [i for j in accept.items() for i in j]
		elif not isinstance(accept, Iterable):
			accept = accept,

		if fuzzy:
			bestValues = (
				section[bestMatch]
				for section in (self[_] for _ in self.sections())
				if (bestMatch := best_match(key, section.keys(), cutoff=0.9))
				   and section.name not in ignoreSection
			)
			matches = (
				scoredMatch
				for bestValueFromSection in bestValues
				if (
					   scoredMatch := scored_get_close_matches(
						   bestValueFromSection, accept, n=1, cutoff=fuzzy_threshold
					   )
				   )
				   and bestValueFromSection not in ignore
				   and (guard is None or guard(bestValueFromSection))
			)
			matches = [i for j in matches for i in j]
			if matches:
				match = matches[0].value
				if "wasMapping" in locals() and match in wasMapping:
					return wasMapping[match]
				return match
			else:
				return default
		return next(
			(
				value
				for section in (self[_] for _ in self.sections())
				if key in section
				   and (value := section[key]) not in ignore
				   and (accept is None or value in accept)
				   and (guard is None or guard(value))
			),
			default,
		)


config = Config()

# search = config.search('locale', fuzzy=True, default=locale.getlocale()[0], accept=locale.locale_alias)
locale.setlocale(locale.LC_ALL, '')

try:
	RADIX_CHAR = locale.nl_langinfo(locale.RADIXCHAR)
	GROUPING_CHAR = locale.nl_langinfo(locale.THOUSEP)
except AttributeError:
	RADIX_CHAR = '.'
	GROUPING_CHAR = ','



from ..base.Decorators import UnitType
from . import Length, Mass, DerivedMeasurement, Volume

__all__ = ['Density']


@UnitType
class Density(DerivedMeasurement, min=0, numerator=Mass, denominator=Volume):
	_id = 'density'

	def __new__(cls, numerator: Mass, denominator: Volume = 1):
		if not isinstance(denominator, Volume) and isinstance(denominator, Length):
			denominator = Volume.compatibleUnits[denominator.unit](denominator)
		return super().__new__(cls, numerator, denominator)

	@property
	def localize(self):
		self.unit
		return super().localize

	@property
	def unit(self):
		return super().unit

	def __class_getitem__(cls, item):
		return super().__class_getitem__(item)



from .mass import Mass
from .metric import *
from .imperial import *

__all__ = ['Mass']

Mass.Microgram = Microgram
Mass.Milligram = Milligram
Mass.Gram = Gram
Mass.Kilogram = Kilogram
Mass.Dram = Dram
Mass.Ounce = Ounce
Mass.Pound = Pound
Mass.Hundredweight = Hundredweight
Mass.Ton = Ton



from ..base import Dimension


class Length(metaclass=Dimension, symbol='L'):
	Millimeter: type
	Centimeter: type
	Decimeter: type
	Meter: type
	Decameter: type
	Hectometer: type
	Kilometer: type

	Line: type
	Inch: type
	Foot: type
	Yard: type
	Mile: type

	@property
	def millimeter(self):
		from . import Millimeter
		return Millimeter(self)
	mm = millimeter

	@property
	def centimeter(self):
		from . import Centimeter
		return Centimeter(self)
	cm = centimeter

	@property
	def meter(self):
		from . import Meter
		return Meter(self)
	m = meter

	@property
	def kilometer(self):
		from . import Kilometer
		return Kilometer(self)
	km = kilometer

	@property
	def line(self):
		from . import Line
		return Line(self)
	ln = line

	@property
	def inch(self):
		from . import Inch
		return Inch(self)

	@property
	def foot(self):
		from . import Foot
		return Foot(self)
	ft = foot

	@property
	def yard(self):
		from . import Yard
		return Yard(self)
	yd = yard

	@property
	def mile(self):
		from . import Mile
		return Mile(self)
	mi = mile

	def __getitem__(self, item):
		item = 'inch' if item == 'in' else item
		super().__getitem__(item)




from ...others import Direction, Voltage, LightningStrike, Humidity
from ...others.light import UVI, Lux, RadiantFlux
from ...time_ import *
from ...pressure import Millibar
from ...derived import Wind, Precipitation
from ...derived.precipitation import Daily as PrecipitationDaily, Hourly as PrecipitationHourly, Minutely as PrecipitationMinutely
from ...temperature import Celsius
from ...length import Kilometer, Meter, Millimeter
from ...various import RSSI
from enum import Flag as _Flag


class SensorStatus(_Flag):
	OK = 0b000000000
	LightningFailed = 0b000000001
	LightningNoise = 0b000000010
	LightningDisturbance = 0b000000100
	PressureFailed = 0b000001000
	TemperatureFailed = 0b000010000
	RHFailed = 0b000100000
	WindFailed = 0b001000000
	PrecipitationFailed = 0b010000000
	LightFailed = 0b100000000


class Debug(_Flag):
	Disabled = 0
	Enabled = 1


class ResetFlags(_Flag):
	BOR = 'Brownout'
	PIN = 'PIN'
	POR = 'Power'
	SFT = 'Software'
	WDG = 'Watchdog'
	WWD = 'WindowWatchdog'
	LPW = 'LowPower'


UDPClasses = {
		'time':                  int,
		'lullSpeed':             'Wind',
		'windSpeed':             'Wind',
		'gustSpeed':             'Wind',
		'speed':                 'Wind',
		'direction':             Direction,
		'windDirection':         Direction,

		'windSampleInterval':    Time.Second,
		'pressure':              Millibar,
		'temperature':           Celsius,
		'humidity':              Humidity,

	'illuminance':             Lux,
	'uvi':                     UVI,
	'irradiance':              RadiantFlux,

	'precipitationRate':       'Precipitation',
	'precipitationDaily':      'PrecipitationDaily',
	'precipitationType':       Precipitation.Type,

	'distance':                Kilometer,
	'lightningLastDistance':   Kilometer,
	'lightning':               LightningStrike,
	'lightningEnergy':         int,

	'battery':                 Voltage,
	'reportInterval':          Time.Minute,
	'Wind':                    (Wind, Meter, Time.Second),
	'Precipitation':           (Precipitation.Hourly, Millimeter, Time.Minute),
	'PrecipitationDaily':      (Precipitation.Daily, Millimeter, Time.Day),

	'deviceSerial':            str,
	'hubSerial':               str,
		'uptime':                Time.Second,
		'firmware':              int,
		'deviceRSSI':            RSSI,
		'hubRSSI':               RSSI,
		'sensorStatus':          SensorStatus,
		'debug':                 Debug,
		'resetFlags':            'resetFlag',
		'resetFlag':             ResetFlags,
}

'''
Sky
0	Time Epoch	Seconds
1	Illuminance	Lux
2	UV	Index
3	Rain amount over previous minute	mm
4	Wind Lull (minimum 3 second sample)	m/s
5	Wind Avg (average over report interval)	m/s
6	Wind Gust (maximum 3 second sample)	m/s
7	Wind Direction	Degrees
8	Battery	Volts
9	Report Interval	Minutes
10	Solar Radiation	W/m^2
11	Local Day Rain Accumulation	mm
12	Precipitation Type	0 = none, 1 = rain, 2 = hail
13	Wind Sample Interval	seconds

Air
0	Time Epoch	Seconds
1	Station Pressure	MB
2	Air Temperature	C
3	Relative Humidity	%
4	Lightning Strike Count
5	Lightning Strike Avg Distance	km
6	Battery
7	Report Interval	Minutes

Tempest
0	Time Epoch	Seconds
1	Wind Lull (minimum 3 second sample)	m/s
2	Wind Avg (average over report interval)	m/s
3	Wind Gust (maximum 3 second sample)	m/s
4	Wind Direction	Degrees
5	Wind Sample Interval	seconds
6	Station Pressure	MB
7	Air Temperature	C
8	Relative Humidity	%
9	Illuminance	Lux
10	UV	Index
11	Solar Radiation	W/m^2
12	Rain amount over previous minute	mm
13	Precipitation Type	0 = none, 1 = rain, 2 = hail, 3 = rain + hail (experimental)
14	Lightning Strike Avg Distance	km
15	Lightning Strike Count
16	Battery	Volts
17	Report Interval	Minutes

'''


class UDP:
	# __slots__ = ['messageTypes', 'Message', 'Wind', 'Event', 'ObservationSet', 'Status', 'Device', 'Hub']

	messageTypes = {
			'air':        'obs_air',
			'sky':        'obs_sky',
			'tempest':    'obs_st',

			'obs_st':     ['time', 'lullSpeed', 'windSpeed', 'gustSpeed', 'windDirection',
			               'windSampleInterval', 'pressure', 'temperature', 'humidity',
			               'illuminance', 'uvi', 'irradiance', 'precipitationRate',
			               'precipitationType', 'lightningLastDistance', 'lightning',
			               'battery', 'reportInterval'],

			'obs_sky':    ['time', 'illuminance', 'uvi', 'precipitationHourlyRaw', 'lullSpeed',
			               'windSpeed', 'gustSpeed', 'windDirection', 'battery', 'reportInterval',
			               'irradiance', 'precipitationDaily', 'precipitationType', 'windSampleInterval'],

			'obs_air':    ['time', 'pressure', 'temperature', 'humidity', 'lightning', 'lightningLastDistance',
			               'battery', 'reportInterval'],

			'wind':       'rapid_wind',
			'rain':       'evt_precip',
			'lightning':  'evt_strike',

			'rapid_wind': ['time', 'speed', 'direction'],

			'evt_strike': ['time', 'distance', 'energy'],

			'evt_precip': ['time']
	}

	Message = {'serial_number': str, 'hub_sn': str, 'type': str}
	Wind = {**Message, 'ob': list}
	Event = {**Message, 'evt': list}
	ObservationSet = {**Message, 'obs': list}

	Status = {'serial_number': str, 'type': str, 'timestamp': int, 'uptime': Time.Second, 'firmware_revision': int, 'rssi': RSSI}
	Device = {**Status, 'hub_sn': str, 'voltage': Voltage, 'hub_rssi': RSSI, 'sensor_status': SensorStatus, 'debug': Debug}
	Hub = {**Status, 'reset_flags': ResetFlags, 'radio_stats': list}



from typing import Type, overload, Tuple

from .config import config

from . import base

SmartFloat = base.SmartFloat
Measurement = base.Measurement
ScalingMeasurement = base.ScalingMeasurement
Dimension = base.Dimension
Dimensionless = base.Dimensionless
NonPlural = base.NonPlural
DerivedMeasurement = base.DerivedMeasurement
Index = base.Index
Quantity = base.Quantity
SystemVariant = base.SystemVariant
FiniteField = base.FiniteField

from .others import (Light, Angle, Percentage, Voltage,
                     LightningStrike, Humidity, Direction,
                     Coverage, Probability)
from .temperature import Temperature
from .length import Length
from .mass import Mass
from .time_ import Time
from .pressure import Pressure
from . import airQuality as AirQuality
from . import digital as Digital
from . import derived

Rate = derived.DistanceOverTime
Wind = derived.Wind
Precipitation = derived.Precipitation
Other = others
Light = others.Light
PartsPer = derived.PartsPer
Volume = derived.Volume
Density = derived.Density
PartsPer = derived.PartsPer


@overload
def auto(value: int | float | int, unit: str) -> Measurement: ...


@overload
def auto(value: str) -> Measurement: ...


@overload
def auto(unit: str) -> Type[Measurement] | Tuple[Type[Measurement], ...]: ...


def auto(*args) -> Measurement | Type[Measurement] | Tuple[Type[Measurement], ...]:
	match args:
		case [int(value) | float(value) | str(value), str(unit)]:
			measurement = float(value)
			unit = Measurement.__findUnitClass__(unit)
			cls = tuple(Measurement.__findUnitClass__(u) for u in unit)

			if len(cls) == 1:
				return cls[0](measurement)
			raise NotImplementedError('Multi-Unit values are not supported yet')

		case [str(value)]:
			number = base.FormatSpec.number.search(value)
			if number is not None:
				number = number.groupdict()['number']

			unit = DerivedMeasurement.__parse_unit__(value.replace(number, ''))
			cls = tuple(Measurement.__findUnitClass__(u) for u in unit)

			if len(cls) == 1:
				return cls[0](float(number))
			raise NotImplementedError('Multi-Unit values are not supported yet')

		case [str(unit)]:
			unit = DerivedMeasurement.__parse_unit__(unit)
			cls = tuple(Measurement.__findUnitClass__(u) for u in unit)
			return cls
		case _:
			raise ValueError


__all__ = ['auto', 'Measurement', 'Temperature', 'Length', 'Mass', 'Time', 'Pressure', 'Other', 'Light', 'PartsPer',
	'Volume', 'Density', 'PartsPer', 'Rate', 'Wind', 'Precipitation', 'AirQuality', 'Voltage', 'Digital', 'config',
	'Precipitation', 'LightningStrike', 'Angle', 'Percentage', 'Humidity', 'Direction', 'Coverage', 'Probability']



from ..base.Decorators import Tiny, Small, Medium, Large
from ..base import Scale, ScalingMeasurement
from ..base import metric
from .length import Length


class MetricLength(ScalingMeasurement, Length, system=metric, baseUnit='Meter'):
	class _Scale(Scale):
		Millimeter = 1
		Centimeter = 10
		Decimeter = 10
		Meter = 10
		Decameter = 10
		Hectometer = 10
		Kilometer = 10
		Base = 'Meter'

	def _foot(self):
		return self.changeScale(self._Scale.Meter) * 3.280839895013123


@Tiny
class Millimeter(MetricLength):
	_unit = 'mm'


@Small
class Centimeter(MetricLength):
	_unit = 'cm'


@Small
class Decimeter(MetricLength):
	_unit = 'dm'


@Medium
class Meter(MetricLength):
	_unit = 'm'


@Medium
class Decameter(MetricLength):
	_unit = 'dam'


@Large
class Hectometer(MetricLength):
	_unit = 'hm'


@Large
class Kilometer(MetricLength):
	_unit = 'km'



from .length import Length
from .metric import *
from .imperial import *

__all__ = ['Length']

Length.Millimeter = Millimeter
Length.Centimeter = Centimeter
Length.Decimeter = Decimeter
Length.Meter = Meter
Length.Decameter = Decameter
Length.Hectometer = Hectometer
Length.Kilometer = Kilometer

Length.Line = Line
Length.Inch = Inch
Length.Foot = Foot
Length.Yard = Yard
Length.Mile = Mile



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['WeatherUnits',
 'WeatherUnits.airQuality',
 'WeatherUnits.base',
 'WeatherUnits.config',
 'WeatherUnits.defaults',
 'WeatherUnits.defaults.WeatherFlow',
 'WeatherUnits.derived',
 'WeatherUnits.digital',
 'WeatherUnits.errors',
 'WeatherUnits.length',
 'WeatherUnits.mass',
 'WeatherUnits.others',
 'WeatherUnits.pressure',
 'WeatherUnits.temperature',
 'WeatherUnits.time_']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'weatherunits',
    'version': '0.6.2',
    'description': 'Easy conversion for common weather units',
    'long_description': "# WeatherUnits️ 🌧\nEasily convert typical weather units from one unit to another with automatic localization.\n\n## Features\n🌈 Effortlessly define or extend your own units\n\n☀️ Change nearly every part of unit display from a config.ini file. No need for additional code.\n\n❄️ Localization can is as simple as adding .localize to a units variable or setting it to automatic\n\n## Requirements\n\n- Python 3.7\n\n## How to use\n\n**Note: The following documentation is currently out of date after some major changes**\n\n### Unit Conversion\n\nConversion is as easy as using the desired unit's name or aliases as either an attribute or subscript\n\n```python\nfrom WeatherUnits.temperature import Fahrenheit\n>>> value = Fahrenheit(32)\n>>> value\n'32º'\n>>> value.withUnit\n'32ºf'\n>>> value['c'].withUnit\n'0ºc'\n>>> value.celcius\n'0º'\n```\n\n## Using a Config file\n\n#### Loading a config file\n```python\nimport WeatherUnits as wu\nwu.config.read('config.ini')\n```\nA template config file can be found in [src/WeatherUnits/config](src/WeatherUnits/config/template.ini).  There are three sections in a valid config file\n1) LocalUnits: This is what defines what units you want to use for localization\n2) UnitDefaults: This section holds the default values that apply to every unit\n3) UnitProperties: Defines properties for individual units or unit systems as a whole\n\n### UnitDefaults\nThe basic rules are:\n\n1) Properties are specified without the leading '_'\n2) Units are specified with their lowercase names or types\n3) Empty values will be read as '' and denote being set\n4) Values without an '=' will be read as None and denote being unset\n\n#### Properties\n- **precision:**  The number of digits allowed to be displayed past the decimal but will not exceed max Example number:\n\n    |example|3.14159m|\n    |-------|--------|\n    |0|3m|\n    |1|3.1m|\n    |2|3.14m|\n    |3|3.141m|\n\n- **max:**  The total number of digits that can be displayed while still showing the decimal.  If shorten is True, numbers would be reduced by factors of 10³\n\n    |example|415.25mm|\n    |-----|-------|\n    |2|0.4k mm|\n    |3|415mm|\n    |4|415.3mm|\n    |5|415.25mm|\n\n- **unit:** Override built in unit string\n- **suffix:** Override built in suffix string._Only used for shortening numbers_\n- **decorator:** Override built in decorator._Only used for º with degrees_\n- **title:** Title to be used by other display programs\n- **exp:** Override exponent.  I can't imagine why this would be used\n- **showUnit:** Show unit True: 5km False: 5\n- **leadingZero:** Display zero before values less than 1 True: 0.1in False: .1in\n- **trailingZero:** Display zero after decimal point to full precision staying under max True: 1.0in False: 1in True: 4.00cm\n- **unitSpacer:** Determines if there is a space between measurements and their units True: 90º f False: 90ºf\n- **kSeparator:**\n\n    |value|example|\n    |----|------|\n    |True|1,000ft|\n    |False|1000ft|\n\n- **shorten:** Shortens values to fit within the max display amount\n\n    |value|example|\n    |----|------|\n    |True|1k s|\n    |False|1000s|\n\n- **slide:** Scale units for the appropriate values. _Not yet implemented_\n\n    |value|example|\n    |----|------|\n    |True|1km|\n    |False|1000m|\n\n- **cardinal:** Show degrees as cardinal direction. **_Shorten_** _decides full vs abbreviated name_\n\n    |value|example|\n    |----|------|\n    |True|180º|\n    |False|South|\n\n- **key:** Key to be used for other programs\n- **sizeHint:** Override generated size hint string. Useful for when you know the expected max string length.\n\n\n## Defining Your Own Unit\n\nThere are currently three types of unit, **Static**, **Scaling**, and **Derived**\n\n### Static Unit\nThe most simple to define. Temperature is an example of a static unit.\nSimilar units of the same are a subclass of the unit type. In the example below,\nFahrenheit, Celsius, and Kelvin are subclasses of Temperature. Convert to other\nsimilar units with a single private function sharing the name of the desired unit.\n\n```python\nfrom WeatherUnits.base import Measurement, NamedType\n\n@NamedType  # Unit types are defined with this decorator.\nclass Temperature(Measurement):  # For static unit types, the class inherits Measurement.\n\n    # Shared unit properties are defined here or in a provided config file.\n    _decorator = 'º'\n\n    # Properties for converting units are defined in the main parent class.\n    @property\n    def fahrenheit(self):\n        # Unless all the subunits are defined in the same file, importing the desired\n        # class locally is currently necessary.\n        return Fahrenheit(self._fahrenheit())\n\n    # Unit abbreviations are defined somewhere below the converting property.\n    f = fahrenheit\n\n    @property\n    def celsius(self):\n        return Celsius(self._celsius())\n    c = celsius\n\nclass Fahrenheit(Temperature):\n    # Individual unit strings are defined as protected class variables\n    # for each unit or within a config file.\n    _unit = 'f'\n\n    # This method is what is called to convert Fahrenheit to Celsius.\n    def _celsius(self):\n        # Since Measurement is a float subclass, math can be done on self.\n        return (self - 32) / 1.8\n\n    def _fahrenheit(self):\n        # Currently, a function needs to be defined for even converting to the\n        # same unit to prevent errors.\n        return self\n\nclass Celsius(Temperature):\n    _unit = 'c'\n\n    def _celsius(self):\n        return self\n\n    def _fahrenheit(self):\n        return (self * 1.8) + 32\n```\n\n### Scaling Unit\n\nScaling units have quite a bit going on.  A scaling unit can be reduced to a single unit through a multiplier, this is the base unit.\nMultiple unit systems can be defined within one scale.  It is also to split systems separate scales as long as there is a function within\neach system that converts from one base unit to the other.\n\n- **SystemVariant:** A class that denotes is not the unit is not within the regular scale\n- **Synonym:** Denotes that a unit is the same as an already defined but uses another name. Synonym classes inherit their identically valued class\n- **Dimension:** Decorator that assigns the decorated class to _unitSystem of all child classes\n- **BaseUnit:** Decorator to define the base unit for the system, this will be used for converting to the non-standard 'SystemVariant'\n- **Scale:** Uses the metaclass EnumMeta to define the scaling factors for a ScalingMeasurement class along with the multipliers for any SystemVariants\nand the base unit\n\nTo define a scaling measurement (one that has multiple units within the same unit system, think Centimeter, Meter, Kilometer)\nthe unit system class inherits ScalingMeasurement.  The main difference from Measurement is the _Scale class defined inside the\nmain parent class.  In the example below, two unit systems (US and SI) are defined in the same scale.  Most US units\nhave an SI basis, so this can be done fairly often.  However, it is sometimes necessary to split them for readability.\n\n#### Unified System\n\n```python\n\nfrom WeatherUnits.base import NamedType, Synonym, ScalingMeasurement\nfrom WeatherUnits.base import Scale, BaseUnit, SystemVariant, Dimension\n\n@NamedType\nclass Pressure(ScalingMeasurement):\n\n    # The scale for the unit system is defined with by a class named _Scale inheriting.\n    class _Scale(Scale):\n        # Every unit is named here starting with the smallest unit and assigned a\n        # scaling multiplier.  Scaling multipliers here must be an integer.\n        Pascal = 1\n        Decapascal = 10\n        Hectopascal = 10\n\n        # A base unit is also defined denoted by Base = class.__name__.\n        # Base value must be a string.\n        Base = 'Pascal'\n\n        # System variants are units that fall outside of the normal scale\n        # The multiplier value for variants must be a float and be relative\n        # to the base unit.  For example, 1 Bar is 1000 Pascals and\n        # 1 MillimeterOfMercury is 1/0.00750062 or 133.3 Pascals.\n        Bar = 1000.\n        Atmosphere = 101325.\n        PoundsPerSquareInch = 6894.757293168\n        MillimeterOfMercury = 1 / 0.00750062\n        InchOfMercury = 1 / 0.00029530\n\n    @property\n    def pascal(self):\n        return Pascal(self)\n\n    @property\n    def hectopascal(self):\n        return Hectopascal(self)\n\n@BaseUnit\nclass Pascal(Pressure):\n    _unit = 'Pa'\n\nclass Hectopascal(Pressure):\n    _unit = 'hPa'\n\nclass Atmosphere(Pascal, SystemVariant):\n    _unit = 'atm'\n\n@Synonym\nclass Millibar(Hectopascal):\n    _unit = 'mBar'\n    _max = 4\n```\n\n#### Split system\n\n```python\nfrom WeatherUnits.base import NamedType, ScalingMeasurement, Dimension, BaseUnit\n\n\n@NamedType\nclass Length(ScalingMeasurement):\n\n    @property\n    def meter(self):\n        return Meter(self)\n    m = meter\n\n    @property\n    def foot(self):\n        return Foot(self)\n    ft = foot\n\n\nclass ImperialLength(Length):\n\n    class _Scale(Scale):\n        Line = 1\n        Inch = 12\n        Foot = 12\n        Yard = 3\n        Mile = 1760\n        Base = 'Foot'\n\n    # Since the multipliers are defined in _Scale, changeScale() uses one\n    # parameter to change the scale within the same system.\n    def _foot(self):\n        return self.changeScale(self._Scale.Foot)\n\n    # This function is called to convert from ImperialLength to MetricLength.  Once its\n    # converted to MetricLength, changeScale() is called within the MetricLength class\n    # to get to the final unit.\n    def _meter(self):\n        return self._foot() * 0.3048\n\n\n@BaseUnit\nclass Foot(ImperialLength):\n    _unit = 'ft'\n\n\n@Dimension\nclass MetricLength(Length):\n\n    class _Scale(Scale):\n        Millimeter = 1\n        Centimeter = 10\n        Decimeter = 10\n        Meter = 10\n        Decameter = 10\n        Hectometer = 10\n        Kilometer = 10\n        Base = 'Meter'\n\n    def _meter(self):\n        return self.changeScale(self._Scale.Meter)\n\n    # As with the pervious class, this method is called to convert from MetricLength\n    # to ImperialLength.\n    def _foot(self):\n        return self._meter() * 3.280839895013123\n\n\n@BaseUnit\nclass Meter(MetricLength):\n    _unit = 'm'\n```\n\n\n### Derived Unit\nDerivedMeasurement is essentially a measurement with multiple units, eg, m/s [meters per second], mph [miles per hour].\nInitiation takes in two Measurements and is stored as the dividend of the two.  Both values are independently stored\nas protected variables.\n\n**Note:**\nThis type still needs quite a lot of work.  Currently, it only has fractional support, and only a single unit can be used for\nboth the numerator and denominator.  Eventually this will be expanded allowing for more complex unit derivatives.\nSupport for the following has not been added:\n - Multiplying two units together, for example a pascal second [Pa·s]\n - Negative exponent units like Hertz [s⁻¹]\n - Multiple units in either the numerator or denominator\n\n```python\nfrom WeatherUnits import Length, Time\nfrom WeatherUnits.base import DerivedMeasurement, NamedType, NamedSubType\n\n\n# Summary of the DerivedMeasurement class\nclass DerivedMeasurement(Measurement):\n    _numerator: Measurement\n    _denominator: Measurement\n\n    def __init__(self, numerator, denominator):\n        self._numerator = numerator\n        self._denominator = denominator\n        Measurement.__init__(self, numerator / denominator)\n\n\n@NamedType\nclass DistanceOverTime(DerivedMeasurement):\n    # Defining the unit types for the derived unit helps keep things clear\n    _numerator: Length\n    _denominator: Time\n\n    # Currently all variations that will be used have to be defined as properties\n    @property\n    def mih(self):\n        converted = DistanceOverTime(self._numerator.mi, self._denominator.hr)\n\n        # Without this override, the unit would be displayed as 'mi/hr'\n        converted._suffix = 'mph'\n        return converted\n\n    @property\n    def ms(self):\n        return DistanceOverTime(self._numerator.m, self._denominator.s)\n\n    @property\n    def kmh(self):\n        return DistanceOverTime(self._numerator.km, self._denominator.hr)\n\n    mph = mih\n\n\n@NamedType\nclass Precipitation(DistanceOverTime):\n    _numerator: Length\n    _denominator: Time\n\n    @property\n    def inh(self):\n        return Precipitation(self._numerator.inch, self._denominator.hr)\n\n    @property\n    def mmh(self):\n        return Precipitation(self._numerator.mm, self._denominator.hr)\n\n\n@NamedType\nclass PrecipitationRate(Precipitation):\n    pass\n\n\n@NamedSubType\nclass Daily(Precipitation):\n\n    def __init__(self, numerator: Length, denominator: int = 1, *args, **kwargs):\n        if isinstance(denominator, int):\n            denominator = Time.Day(denominator)\n        Precipitation.__init__(self, numerator, denominator, *args, **kwargs)\n\n```",
    'author': 'noblecloud',
    'author_email': 'git@noblecloud.io',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/noblecloud/WeatherUnits',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from fractions import Fraction
from functools import cached_property
from typing import Optional, Tuple, Mapping

from ..base import Dimensionless, NonPlural, Measurement, Quantity, FiniteField, UnitType
from ..utils import getFrom
from . import light

Light = light.Light

__all__ = ['Light', 'Angle', 'Direction', 'Humidity', 'Voltage', 'LightningStrike', 'Probability', 'Coverage', 'Percentage']


@UnitType
class Percentage(Dimensionless, NonPlural):
	_decorator = '%'
	_max = 3
	_precision = 2

	def __new__(cls, value, isPercentage: bool = None, *args, **kwargs):
		if isPercentage is None:
			if value > cls._limits[1] or isinstance(value, int) and value != 1:
				value /= 100
		elif isPercentage:
			value /= 100

		return super(Percentage, cls).__new__(cls, value, *args, **kwargs)

	def __format__(self, format_spec: str) -> str:
		return super().__format__(format_spec)

	def __int__(self) -> int:
		return int(float(self)*100)

	# @property
	# def defaultFormatParams(self):
	# 	sup = super().defaultFormatParams
	# 	sup['type'] = '%'
	# 	return sup

	@classmethod
	def fromFloat(cls, value: float) -> 'Percentage':
		return cls(value, isPercentage=False, limits=False)

	@property
	def formatValue(self) -> float:
		return float(self)*100


class Humidity(Percentage, limits=(0.0, 1.0)):
	_decorator = '%'
	_id = '%h'


class Probability(Percentage, limits=(0.0, 1.0)):
	_decorator = '%'
	_id = '%p'
	_fraction: Optional[Fraction]
	_denominatorLimit: int = 10

	@property
	def fraction(self) -> Fraction:
		if (fraction := getattr(self, '_fraction', None)) is None:
			self._fraction = Fraction(self).limit_denominator(self._denominatorLimit)
		return fraction


class Coverage(Percentage, limits=(0.0, 1.0)):
	_decorator = '%'
	_id = '%c'


class BatteryPercentage(Percentage, limits=(0.0, 1.0)):
	_voltageLimits: Optional[Tuple[float, float]]
	_decorator = '%'
	_id = '%bat'

	def __init__(self, value, voltageLimits: Optional[Tuple[float, float]] = None, *args, **kwargs):
		self._voltageLimits = voltageLimits
		super(BatteryPercentage, self).__init__(value, *args, **kwargs)


@UnitType
class Angle(Dimensionless):
	_precision = 0
	_decorator = 'º'
	_shorten = True


class Direction(Angle, FiniteField, limits=(0, 360)):
	_cardinal = True
	_decorator = 'º'
	_id = '°d'
	_max = 3

	@cached_property
	def cardinal(self) -> 'Cardinal':
		return Cardinal(self)

	@property
	def angle(self) -> Angle:
		return Angle(self)

	@property
	def decoratedInt(self) -> str:
		return super()._string(forceUnit=False, asInt=True)

	@property
	def __cardinalFormat(self):
		return '{value}'

	@property
	def __valueFormat(self):
		return "{value}{decorator}"

	@property
	def defaultFormat(self) -> str:
		if self._cardinal:
			return self.__cardinalFormat
		return self.__valueFormat

	@property
	def defaultFormatParams(self):
		return {
			'cardinal':     self._cardinal if self._cardinal else False,
			'showCardinal': self._cardinal,
			**super().defaultFormatParams
		}

	@property
	def properties(self):
		measurement = super().properties
		return {'cardinal': self.cardinal, 'showCardinal': self._cardinal, **measurement, }

	def __repr_value__(self) -> str:
		return f'{self: cardinal: False}'

	def __format_value__(self, params: Mapping) -> str:
		showCardinal = getFrom(('showCardinal', 'cardinal'), *params.maps, default=self._cardinal, expectedType=(bool, str))
		if showCardinal:
			return self.cardinal.__format_value__(params)
		return super().__format_value__(params)

	def __format_template__(self, params: Mapping) -> str:
		template = super().__format_template__(params)
		showCardinal = getFrom(('showCardinal', 'cardinal'), *params.maps, default=self._cardinal, expectedType=(bool, str))
		if showCardinal and '{cardinal}' not in template:
			template = template.replace('{decorator}', '')
		return template


class Cardinal:
	__slots__ = '__direction'
	__dirsAbbrv = ['N', 'NNE', 'NE', 'ENE', 'E', 'ESE', 'SE', 'SSE', 'S', 'SSW', 'SW', 'WSW', 'W', 'WNW', 'NW', 'NNW']
	__dirsFull = ['North', 'North Northeast', 'Northeast', 'East Northeast', 'East', 'East Southeast', 'Southeast',
		'South Southeast', 'South', 'South Southwest', 'Southwest', 'West Southwest', 'West',
		'West Northwest', 'Northwest', 'North Northwest']
	__direction: Direction
	direction: Direction

	def __init__(self, direction: Direction):
		self.__direction = direction

	def __str__(self):
		if self.shorten:
			return self.abbrivated
		else:
			return self.full

	def __format_value__(self, params: Mapping) -> str:
		if self.shorten:
			return self.abbrivated
		else:
			return self.full

	def __repr__(self):
		return f'Cardinal({self!s} {self.direction: cardinal: False})'

	@property
	def direction(self) -> Direction:
		return self.__direction

	@property
	def __shortIndex(self):
		return round(self.direction/45)%8*2

	@property
	def __longIndex(self):
		return round(self.direction/22.5)%16

	@property
	def abbrivated(self):
		if self.direction.max > 2:
			return self.threeLetter
		return self.twoLetter

	@property
	def shortened(self):
		if self.direction.max < 5:
			return self.abbrivated
		return self.singleWord

	@property
	def twoLetter(self):
		return self.__dirsAbbrv[self.__shortIndex]

	@property
	def threeLetter(self):
		return self.__dirsAbbrv[self.__longIndex]

	@property
	def singleWord(self):
		return self.__dirsFull[self.__shortIndex]

	@property
	def text(self):
		if self.direction.max < 7:
			return self.singleWord
		return self.full

	@property
	def full(self):
		return self.__dirsFull[self.__longIndex]

	@property
	def shorten(self):
		return self.direction.shorten

	@shorten.setter
	def shorten(self, value: bool):
		self.direction.shorten = value


class Voltage(Measurement):
	_max = 3
	_precision = 2
	_unit = 'v'


class LightningStrike(Quantity, altName='Strike'):
	...


Percentage.Humidity = Humidity
Percentage.Probability = Probability
Percentage.Coverage = Coverage
Percentage.BatteryPercentage = BatteryPercentage

Angle.Direction = Direction
Direction.Cardinal = Cardinal
Angle.Cardinal = Cardinal
